(**
This file is part of the IE library

Copyright (C) El Ouraoui, Mayero, Mazza

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
COPYING file for more details.
*)


Require Export Arith.
Require Export Compare_dec.
Require Export Relations.
Require Export Omega.



(** Les termes *)

Inductive term : Set :=
| TVar : nat -> term
| TAbs : term -> term
| TApp : term -> term -> term
| TSub : term -> term -> term.
Notation "t [ u ]" := (TSub t u)(at level 70).
Notation "\" := (TAbs ) (at level 66).
Notation "u @ v" := (TApp u v) (at level 66).

(**  Mise à jours des indices de de Bruijn *)

Fixpoint lift_rec n t {struct t} : nat -> term :=
  fun k =>
    match t with
      | TVar i  =>
        match le_gt_dec k i with
          | left _ => TVar (n + i)
          | right _ => TVar i
        end
      | \u => \(lift_rec n u (S k))
      | u @ v => (lift_rec n u k) @ (lift_rec n v k)
      | u [v]  => (lift_rec n u (S k))[(lift_rec n v k)]
    end.

(*  definition de lift *)

Definition lift n t := lift_rec n t 0.
Definition shift t := lift 1 t.


(** Définition de la substitution explicite M{k/N} *)

Fixpoint subst_rec u t {struct t} : nat -> term :=
  fun k =>
    match t with
      | TVar i =>
        (* si i=k then N else TVar i *)
        match eq_nat_dec k i with  
          | left _ => u
          | right _ => TVar i
        end
      | \v => \(subst_rec (shift u) v (S k))
      | w@v => (subst_rec u w k)@(subst_rec u v k) 
      | w[v] => (subst_rec (shift u) w (S k))[(subst_rec u v k)]
    end.

Definition subs_recbis u x t:=  subst_rec u t x.
Notation " t { x := u } " := (subs_recbis u x t)(at level 0).


(*Definition ls := exists u :term,  lsa_def u 0.*)

Lemma subst_nothing' :
  forall u u',  forall j n, n > 0 -> 
   (lift_rec n u j) { j := u'} = (lift_rec n u j). 
Proof.
  induction u.
  -simpl.  intros. case (le_gt_dec j n).
   *simpl. case (eq_nat_dec (j) (n0+n)).
    +intros. omega.
    +intros;auto.
     *simpl. case (eq_nat_dec j (n)).
      +intros;omega.
      +intros;auto.
  -intros. simpl. rewrite (IHu (lift_rec 1 u' 0) (S j)); auto.
  -intros. simpl. rewrite (IHu1 (u') (j)); auto. simpl. rewrite (IHu2 (u') (j)); auto.
  -intros. simpl. rewrite (IHu1 (lift_rec 1 u' 0) (S j)); auto. simpl. rewrite (IHu2 (u') (j)); auto.
Qed.

Lemma subst_nothing :
  forall u u',
   (shift u) { 0 := u'} = (shift u). 
Proof.
  intros. unfold shift in *. unfold lift. apply subst_nothing';auto with arith.
Qed.
(* lemme d'usage elementaire sur lift ==> lift 0 M = M la preuve est simple par induction sur M*)

Lemma lift_rec0 : forall t k, lift_rec 0 t k = t.
  simple induction t; simpl in *; intros; auto with arith.
  elim (le_gt_dec k n); auto with arith.

  rewrite H;auto with arith.
  rewrite H ; rewrite H0 ; auto with arith .
  rewrite H; rewrite H0; auto with arith.
Qed.


Lemma lift0 : forall t, lift 0 t = t.
  intros; unfold lift in *.
  apply lift_rec0; auto with arith.
Qed.


(* 

Definition du predicat disant qu'aucune variable k n'est libre dans t 
ps: il faut vraiment que je change le nom de ce predicat...

 *)

Fixpoint is_free k t :=
  match t with
    | TVar j => k <> j
    | \u => is_free (S k) u 
    | u@v => is_free k u /\ is_free k v 
    | u[v] => is_free (S k) u /\ is_free k v
  end.

(** Quelques lemmes élémentaires concernant lift et subst *)


Lemma lift_var_ge :
  forall k n p, p <= n -> lift_rec k (TVar n) p = TVar (k + n).
  intros; simpl in *.
  elim (le_gt_dec p n); auto with arith.
  intro; absurd (p <= n); auto with arith.
Qed.


Lemma lift_var_lt : forall k n p, p > n -> lift_rec k (TVar n) p = TVar n.
  intros; simpl in *.
  elim (le_gt_dec p n); auto with arith.
  intro; absurd (p <= n); auto with arith.
Qed.


Lemma subst_var_diff : forall u n k, k <> n -> (TVar n){k := u} = TVar n.
  simpl in *; intros.
  elim (eq_nat_dec k n); [ intro a | intro b; auto with arith].
  contradict a.
  auto.
Qed.

(** Lemme de simplification du lift *)

Lemma simpl_lift_rec :
  forall t n k p i,
    i <= k + n ->
    k <= i -> lift_rec p (lift_rec n t k) i = lift_rec (p + n) t k.
  simple induction t; simpl in *; intros; auto with arith.
  elim (le_gt_dec k n); intros.
  rewrite lift_var_ge; auto with arith.

  rewrite plus_comm; apply le_trans with (k + n0);
  auto with core arith sets.

  rewrite lift_var_lt; auto with arith.
  apply le_gt_trans with k; auto with arith.

  rewrite H; auto with arith;simpl in *;
  auto with arith.

  rewrite H; auto with arith; rewrite H0; simpl in *;
  auto with arith.
  rewrite H; auto with arith.
  rewrite H0; simpl in *;
  auto with arith.
  omega.
Qed.

Lemma simpl_lift : forall t n, lift (S n) t = shift (lift n t).
  intros; unfold shift ; unfold lift in *.
  rewrite simpl_lift_rec; auto with arith.
Qed.


(** Lemme de commutativité des indices pour lift *)


Lemma lift_commut: forall a k i j p, k <= p -> 
                                      lift_rec i (lift_rec j a p) k = lift_rec j (lift_rec i a k) (i + p).
Proof.
  induction a.

  (* var case *)
  intros. simpl. elim (le_gt_dec p n). elim (le_gt_dec k n).
  intros. simpl. elim (le_gt_dec k (j+n)). elim (le_gt_dec (i + p) (i + n)).
  intros. replace (i + (j + n)) with (j + (i + n)) by omega;auto.
  intros. contradict a1. omega.
  intros. elim (le_gt_dec (i + p) (i + n)).
  intros. contradict a1. omega.
  intros. contradict a0. omega.
  intros. simpl. elim (le_gt_dec k (j + n) ). elim (le_gt_dec (i + p) n).
  intros. contradict a1. omega.
  intros. contradict a0. omega.
  elim (le_gt_dec (i + p) n). intros;reflexivity.
  intros;contradict a;omega. intros.
  elim (le_gt_dec k n).
  intros. simpl. elim (le_gt_dec k n). elim (le_gt_dec (i + p) (i + n)).
  intros. contradict a1. omega.
  intro;reflexivity. elim (le_gt_dec (i + p) (i + n)).
  intros. contradict a0. omega.
  intros. contradict a.  omega.
  intros. simpl. elim (le_gt_dec k n). elim (le_gt_dec (i + p) (n)).
  intros. contradict a0. omega.
  intros. contradict a. omega. elim (le_gt_dec (i + p) n).
  intros. contradict a. omega.
  intros;reflexivity.
  
  (*abs *)
  simpl. intros. f_equal. rewrite IHa. replace (i + S p)  with (S (i + p)) by omega;auto.
  omega.

  (*app*)
  simpl. intros. f_equal. rewrite IHa1. auto. omega.
  rewrite IHa2;auto.
  
  (*sub*)
  simpl. intros. f_equal.
  rewrite IHa1. replace (i + S p)  with (S (i + p)) by omega;auto.
  omega. rewrite IHa2;auto.
Qed.



Lemma shift_commut: forall u j, lift j (shift u) = shift (lift j u).
Proof.
  intros.
  unfold shift in *.
  unfold lift in *.
  rewrite lift_commut;auto with arith.
  rewrite simpl_lift_rec in *;auto with arith.
  rewrite simpl_lift_rec;auto with arith;omega.
  omega.
Qed.

(*

 Encore un lemme de commutativite du lift un peu specifique 

 *)

Lemma lift_commut_of_succ:
  forall u n k,
    lift_rec n (lift_rec 1 u k) (k+1) = lift_rec 1 (lift_rec n u k) k.
Proof.
  intro u.
  induction u.
  simpl.
  intros.
  elim (le_gt_dec k n).
  intros.
  rewrite lift_var_ge.
  rewrite lift_var_ge.
  replace (1+(n0 + n)) with ((n0 + S n)) by omega;reflexivity.
  omega.
  omega.
  intros.
  rewrite lift_var_lt.
  rewrite lift_var_lt.
  reflexivity.
  auto.
  omega.
  
  (* cas abstraction *)
  simpl.
  intros.
  f_equal.
  replace (S (k + 1)) with (S k + 1) by omega.
  rewrite (IHu n (S k));reflexivity.
  
  (* cas application *)
  simpl;f_equal;intros.
  rewrite IHu1.
  rewrite IHu2.
  auto.
  
  (* cas sub *)
  simpl;f_equal;intros.
  replace (S (k + 1)) with (S k + 1) by omega.
  rewrite (IHu1 n (S k)).
  rewrite IHu2.
  auto.
Qed.
   
   
(*

Pareil commutativite entre lift et substitution lemme assez utile 

 *)


Lemma commut_lift_subst_rec :
  forall t u n k h,
    k <= h ->
    lift_rec n (t {h := u}) k = (lift_rec n t k) {(n+h) := (lift_rec n u k)}.
  simple induction t; intros; auto with arith.
  simpl.
  elim ( eq_nat_dec h n).
  elim (le_gt_dec k n).
  intros.
  simpl.
  elim (eq_nat_dec (n0 + h) (n0 + n)).
  intros;auto.
  intros.
  contradict b.
  auto.
  simpl.
  elim (eq_nat_dec (n0 + h) n).
  auto.
  intros.
  contradict b0.
  omega.
  elim ( le_gt_dec k n).
  simpl.
  elim (le_gt_dec k n).
  elim (eq_nat_dec (n0 + h) (n0 + n)).
  intros.
  contradict a.
  omega.
  intros.
  auto.
  elim (eq_nat_dec (n0 + h) (n0 + n) ).
  intros.
  contradict b.
  omega.
  intros.
  contradict a.
  omega.
  simpl.
  elim (le_gt_dec k n).
  elim (eq_nat_dec (n0 + h) ( n) ).
  intros.
  contradict b.
  omega.
  intros.
  contradict b.
  omega.
  elim (eq_nat_dec (n0 + h) ( n) ).
  intros.
  contradict b.
  omega.
  intros;auto.

  (* cas abstraction *)
  simpl.
  rewrite H;auto with arith.
  symmetry.
  unfold shift in *. unfold lift in *.
  rewrite lift_commut;auto with arith.
  replace (1+k) with (S k) by omega.
  replace (n + S h) with (S (n+h)) by omega.
  auto.

  (* cas application *)
  simpl.
  rewrite H0;auto with arith.
  rewrite H;auto with arith.

  (* cas abstraction *)
  simpl.
  rewrite H;auto with arith.
  rewrite H0;auto with arith.
  symmetry.
  unfold shift in *. unfold lift in *.
  rewrite lift_commut;auto with arith.
  replace (1+k) with (S k) by omega.
  replace (n + S h) with (S (n+h)) by omega.
  auto.


Qed.

(** Commutativité entre lift et substitution *)

Lemma commut_lift_subst :
  forall t u k, (shift t) {(S k) := shift u} = shift t {k := u}.
  intros.
  unfold  shift in *.
  unfold lift in *.
  rewrite commut_lift_subst_rec; auto with arith.
Qed.

(** Lemmes pour le prédicat gc de la réduction *)

Lemma substFree_eq:
  forall t u x, is_free x t -> t{x := u} = t.
Proof.
  induction t.
  intros x u H.
  rewrite subst_var_diff;auto.
  intros;simpl.
  rewrite (IHt (shift u) (S x));auto.
  intros;simpl.
  simpl in H.
  destruct H.
  rewrite (IHt2 u x);auto.
  rewrite (IHt1 u x);auto.
  intros.
  simpl in H.
  destruct H.
  simpl.
  rewrite (IHt2 u x);auto.
  rewrite (IHt1 (shift u) (S x));auto.
Qed.



Fixpoint  num_occ t n :nat :=
  match t with 
    |TVar x => if n =? x then 1
               else 0
    |\u   =>
     num_occ u (S n) 
    |u @ v =>
     num_occ u n + num_occ v n
    | u [v]   =>
     num_occ u (S n) + num_occ v n
  end.

Lemma var_occ :
  forall n i,
    num_occ (TVar n) i = 0 -> n <> i.
Proof.
  intros. case (Nat.eqb_spec i n).
  -intros;rewrite <- e in *; simpl in *;revert H; case (Nat.eqb_spec);intros;omega. 
  -auto.
Qed.
  
Lemma no_occ_of_x :
  forall t u u',forall i,
    num_occ t i = 0 -> t{i := u} = t{i := u'}.
Proof.
  induction t.
  -intros. apply var_occ in H. rewrite !  subst_var_diff  ;auto.  
  -simpl. intros. f_equal. rewrite (IHt (shift u) (shift u')) ;  auto.
  -simpl. intros. f_equal. rewrite (IHt1 u u') ;  auto with arith. omega.
   *simpl. intros.  apply (IHt2 u u' ) ;  auto with arith. omega.
  -simpl. intros. f_equal.
   rewrite (IHt1 (shift u) (shift u') (S i)) ;  auto with arith. omega.
   *simpl. intros. 
   rewrite (IHt2 u u' i) ;  auto with arith. omega.
Qed.

Lemma free_lift :
  forall u:term,  forall n k j:nat,
    0 <= j ->  j <= k < n+j ->
    num_occ (lift_rec n u j) k = 0.
Proof.
   induction u.
  -intros; simpl in H. simpl.  case le_gt_dec.
   *simpl. case Nat.eqb_spec;auto.
    intros. omega.  
   *intros. simpl. case Nat.eqb_spec.
    intros. omega. auto.    
  -intros. simpl. f_equal.
   rewrite IHu;auto; omega.
  -intros.  simpl. rewrite IHu1;auto with arith. rewrite IHu2;auto with arith;omega;omega. 
  -intros.  simpl;rewrite IHu1;auto with arith. rewrite IHu2; auto with arith; omega;omega. omega.
Qed.  